"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import abc
import flwr.proto.grpcadapter_pb2
import grpc

class GrpcAdapterStub:
    def __init__(self, channel: grpc.Channel) -> None: ...
    SendReceive: grpc.UnaryUnaryMultiCallable[
        flwr.proto.grpcadapter_pb2.MessageContainer,
        flwr.proto.grpcadapter_pb2.MessageContainer]


class GrpcAdapterServicer(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def SendReceive(self,
        request: flwr.proto.grpcadapter_pb2.MessageContainer,
        context: grpc.ServicerContext,
    ) -> flwr.proto.grpcadapter_pb2.MessageContainer: ...


def add_GrpcAdapterServicer_to_server(servicer: GrpcAdapterServicer, server: grpc.Server) -> None: ...
